;====================================================================
; Gigabyte GA-7DPXDW-P custom option ROM to fix bluescreens on ATI X800 (and newer)
; v0.1
; author: Tzk
; 26 Feb 2021
; thanks to tictac, pinczakko, polygon @ rebelshaven forums
; Infrared @ hwbot and digitalbath @ Hardwareluxx
;====================================================================
; Version History
; v0.1: initial release
;====================================================================
use16 ; 16bit mode
ROM_SIZE_IN_BLOCK = 1 ; 1 means ROM size is 1 block (512 bytes)
ROM_SIZE_IN_BYTE = ROM_SIZE_IN_BLOCK * 512 ; number of 512 byte blocks
ROMStart:
db 0x055, 0x0AA ; ROM Header 55,AA -> Bootable rom
db (ROMEnd - ROMStart)/512 ; ROM Size in 512byte
jmp MAIN ; jump to main routine
db 0 ; checksum, to be filled in later
times (256)-($-$$) db 0 ; locate Main routine at 100h
MAIN:
pushfd
push eax
push ebx
push ecx
push dx                    ; save all registers for later restore
;
;====================================================================
;
; bus-device-function registers are devided into 32bit segments.
; We got 256 byte in total per function which corresponds to 64 groups with 32bit (2048 bits).
; We need to choose a 32bit segment, mask the undesired data, set our custom value and write to register.
; example:
; we want to write to Bus 0 Device 0 Function 2 and set Offset 65h
; 65h is in the group "64". Group 60 is Offset 60 to 63, group 64 is offset 64 to 67
; The structure of the 32bit address is:  enable bit -  bus - dev - func - register - 00h
; Your resulting address is: 080000064h
; This gives us the data from offset 64 to 67 -> xx65xxxx
; with the highest offset first: 67-66-65-64
;
; now we need to mask relevant bits and discard them ("and" command)
; afterwards we add new desired bits to the data ("or" command)
;
;====================================================================
;
; macro to set PCI register
;
; this macro is used to apply our custom data. It sets the pci register we hand over.
; it takes 3 variables:
; 1. pci register we want to modify
; 2. data we want to apply
; 3. mask we need to discard the old values inside the register
;
; the mask is needed to leave all other data in the register untouched,
; as nvidia might store multiple settings inside a single PCI register
; Usually the mask is the inverse of the data we want to set.
;
; example call of the macro:
; writeregister 080000464h, 011001100h, 000FF00FFh
;
; this reads Bus 0 Device 0 Function 4 (here: Memcontroller) register 64 to 67
; 011001100h is the data we want to set. This sets register 65 and 67. Note that the highest byte is first.
; 000FF00FFh is the mask we apply. Setting 0 will discard the old value while F will keep it
; as we AND mask and data later, before OR applies our custom values to the now empty registers
;
; we basically repeat this for every pci register we want to write.
;
; data and adresses are read/sent trough special port register 0CF8h (adress) and 0CFCh (data)
; So we need to:
; 1. set address register and sent to adress port (0CF8h)
; 2. set data port register and sent to data port (0CFCh)
; 3. get data from data port
;
;====================================================================
;
;B0:D0:F0:4C register
;we need to set bit 1 and 2 to 1b
;
; 08000004Ch
;
;bus number                    device number           function number       register number
;31 30 29 28 27 26 25 24 | 23 22 21 20 19 18 17 16 | 15 14 13 12 11 10 9 8 | 7 6 5 4 3 2 1 0|
; 1  0  0  0  0  0  0  0    0  0  0  0  0  0  0  0    0  0  0  0  0  0 0 0   0 0 1 0 0 1 1 0
;
;mask:
; 0  0  0  0  0  0  0  0 |  0  0  0  0  0  0  0  0 |  0  0  0  0  0  0 0 0 | 0 0 0 0 0 1 1 0|

mov eax,08000004Ch  ; move address for 32bit register offset
mov dx,0CF8h        ; pci register address port
out dx,eax          ; set port we want to read
mov dx,0CFCh        ; pci register data port
in eax,dx           ; read register values
and eax,0FFFFFFF9h  ; mask for data to remove pre-set value, usually the inverse of the data we set
or eax,000000006h   ; add custom value to old register data
out dx,eax          ; write new custom register value

;====================================================================
;
;B0:D0:F0:84 register
;we need to set bit 23 to 1b and bit 3 to 0b
;
; 080000084h
;
;bus number                    device number           function number       register number
;31 30 29 28 27 26 25 24 | 23 22 21 20 19 18 17 16 | 15 14 13 12 11 10 9 8 | 7 6 5 4 3 2 1 0|
; 0  0  0  0  0  0  0  0 |  0  0  0  0  0  0  1  1 |  0  0  0  1  0  0 0 0 | 1 0 0 1 0 1 1 1|
; mask:
; 1  1  1  1  1  1  1  1    0  1  1  1  1  1  1  1    1  1  1  1  1  1 1 1   1 1 1 1 0 1 1 1
;    FF                         7F                              FF                F7
;
mov eax,080000084h  ; move address for 32bit register offset
mov dx,0CF8h        ; pci register address port
out dx,eax          ; set port we want to read
mov dx,0CFCh        ; pci register data port
in eax,dx           ; read register values
and eax,0FF7FFFF7h  ; mask for data to remove pre-set value, usually the inverse of the data we set
or eax,000800000h   ; add custom value to old register data
out dx,eax          ; write new custom register value
;
;====================================================================
;
pop dx
pop ecx
pop ebx
pop eax
popfd
retf ; return far to system bios routine
times (ROM_SIZE_IN_BYTE-$) db 0 ; use 00h as the padding bytes until we reach the ROM size
; The last byte (512th) will be the patch_byte for the checksum
; patch_byte is calculated and automagically inserted below
PREV_CHKSUM = 0
repeat $
load CHKSUM byte from %-1
CHKSUM = (PREV_CHKSUM + CHKSUM) mod 0x100
PREV_CHKSUM = CHKSUM
end repeat
store byte (0x100 - CHKSUM) at ($-1) ; store the patch_byte
ROMEnd:
